// Sudokus con Clase
// Resolucin de sudokus con tcnicas reales
// (C) 2022 Salvador Pozo
// Con Clase https://conclase.net

#include <fstream>
#include <stdio.h>
#include <windows.h>

char tablero[] = {
    "074006200000400500008000000080004010000257060003800000000300020601000900000509000"
};

// Funcin para mostrar texto en pantalla de consola
void Texto(int f, int c, const char *texto, WORD color) {
    COORD pos;
    DWORD escritos;

    pos.X = c;
    pos.Y = f;
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), color);
    SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE), pos);
    WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), texto, strlen(texto), &escritos, NULL);
}

void LogPantalla(char *cad) {
    HANDLE hOut;
    CONSOLE_SCREEN_BUFFER_INFO csbInfo;
    SMALL_RECT sr_ScrollRect = { 0, 42, 50, 52 };// 42, 0, 52, 50
    SMALL_RECT sr_ClipRect = sr_ScrollRect;
    CHAR_INFO chInfo;
    COORD pos = {0, 41};

    chInfo.Attributes = 15;
    chInfo.Char.AsciiChar = ' ';
    hOut = GetStdHandle(STD_OUTPUT_HANDLE);
    GetConsoleScreenBufferInfo(hOut, &csbInfo);
    ScrollConsoleScreenBuffer(hOut, &sr_ScrollRect, &sr_ClipRect, pos, &chInfo);
    Texto(51, 0, cad, 15);
}

// Pausa hasta que se pulse intro
void Pausa() {
    INPUT_RECORD cad[100];
    DWORD leido;

    ReadConsoleInput(GetStdHandle(STD_INPUT_HANDLE), cad, 100, &leido);
    ReadConsoleInput(GetStdHandle(STD_INPUT_HANDLE), cad, 100, &leido);
}

class Unico {
    private:
    int casilla;
    int valor;
    public:
    Unico(int c, int v) : casilla(c), valor(v) {}
    int Casilla() { return casilla; }
    int Valor() { return valor; }
};

class Casilla {
    private:
    int fila, columna, caja, pos;
    int valor;
    bool posible;
    bool inicial;
    bool resaltar;
    bool lapiz[9];

    public:
    Casilla(int f, int c, int v, bool ini);
    int Valor() const { return valor; }
    void Asignar(int v);
    int Fila() const { return fila; }
    int Columna() const { return columna; }
    int Caja() const { return caja; }
    int PosEnCaja() const { return pos; }
    bool Posible() const { return posible; }
    void Resaltar(bool r=true) { resaltar = r; }
    void MarcaPosible( bool p=true ) { posible = p; }
    void MarcaNoPosible() { posible = false; }
    bool MarcaLapiz(int v);
    bool DesmarcaLapiz(int v);
    bool Lapiz(int v) { return lapiz[v-1]; }
    void Mostrar();
};

// Declaracin forward
class Tablero;

class Fila {
    private:
    Casilla *cas[9];

    public:
    Fila() { for(int i=0; i < 9; i++) cas[i] = NULL; }
    void Iniciar(Tablero &t, int c);
    bool Presente(int n);
    Casilla *Cas(int i) { return cas[i]; }
    //
    void Criba();
    void CribaMenosCaja(int c);
    Unico CandidatoUnico();
    int UnicoPosible();
};

class Columna {
    private:
    Casilla *cas[9];

    public:
    Columna() { for(int i=0; i < 9; i++) cas[i] = NULL; }
    void Iniciar(Tablero &t, int f);
    bool Presente(int n);
    Casilla *Cas(int i) { return cas[i]; }
    //
    void Criba();
    void CribaMenosCaja(int f);
    Unico CandidatoUnico();
    int UnicoPosible();
};

class Caja {
    private:
    Casilla *cas[9];
    int punteroFila[3][9];
    int punteroColumna[3][9];

    public:
    Caja();
    void Iniciar(Tablero &t, int c);
    bool Presente(int n);
    Casilla *Cas(int i) { return cas[i]; }
    //
    void Criba();
    int CuentaPosibles();
    int PrimeroPosible();
    int SegundoPosible(int p);
    int TerceroPosible(int t);
    bool AsignaPunteroFila(int f, int v, int n);
    bool AsignaPunteroColumna(int c, int v, int n);
    bool PunteroFila(int f, int v) const { return punteroFila[f][v-1]; }
    bool PunteroColumna(int c, int v) const { return punteroColumna[c][v-1]; }
    void AsignarLapizUnicos();
    Unico CandidatoUnico();
};

class Tablero {
    private:
    Casilla casilla[9][9];
    Fila fila[9];
    Columna columna[9];
    Caja caja[9];
    int cuenta;
    std::ofstream flog;

    public:
    Tablero(char* t);
    ~Tablero() { flog.close(); }
    void Log(char *cad);
    Casilla &GetCasilla(int f, int c) { return casilla[f][c]; }
    const int Cuenta() { return cuenta; }
    void Mostrar();
    bool Soluciones();
    bool Valido(int f, int c, int v);
    void Asignar(int f, int c, int v);
    // Resolucin:
    void LimpiaResaltes();
    void LimpiaCriba();
    void Criba(int v);
    bool ProcesarCriba(int v);
    bool CandidatoUnico();
    bool ProcesarCribaSolitarios(int v);
};

// Definicin de clase Casilla
Casilla::Casilla(int f=0, int c=0, int v=0, bool ini=true) : fila(f), columna(c), valor(v), posible(false), inicial(ini), resaltar(false) {
    for(int i=0; i<9; i++) lapiz[i] = false;
}

void Casilla::Asignar(int v) {
    valor = v;
    for(int v=1; v<=9; v++) DesmarcaLapiz(v);
}

bool Casilla::MarcaLapiz(int v) {
    if(!lapiz[v-1]) {
        lapiz[v-1] = true;
        return true;
    }
    return false;
}

bool Casilla::DesmarcaLapiz(int v) {
    if(lapiz[v-1]) {
        lapiz[v-1] = false;
        return true;
    }
    return false;
}

void Casilla::Mostrar() {
    char C[2]=" ";
    int v;
    WORD color=0;

    if(!posible)
        color = BACKGROUND_GREEN;// | BACKGROUND_INTENSITY

    if(resaltar)
        color = BACKGROUND_BLUE;// | BACKGROUND_INTENSITY;

    if(valor) {
        color |= FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN;
        if(inicial) color |= FOREGROUND_INTENSITY;
    } else {
        color |= FOREGROUND_INTENSITY | FOREGROUND_RED;
    }
    v=1;
    for(int i=1; i<=3; i++) {
        for(int j=1; j<=3; j++) {
            if(Lapiz(v)) C[0] = '0'+v; else C[0] = ' ';
            Texto(fila*4+i, columna*4+j, C, color);
            v++;
        }
    }

    if(valor) {
        C[0] = '0'+Valor();
        Texto(fila*4+2, columna*4+2, C, color);
    }
}

// Definicin de clase Fila
void Fila::Iniciar(Tablero &t, int f) {
    for(int i = 0; i < 9; i++) cas[i] = &t.GetCasilla(f, i);
}

bool Fila::Presente(int n) {
    for(int i=0; i<9; i++) if(cas[i]->Valor() == n) return true;
    return false;
}

void Fila::Criba() {
    for(int i=0; i<9; i++) cas[i]->MarcaNoPosible();
}

void Fila::CribaMenosCaja(int c) {
    for(int i=0; i<9; i++) if(i/3 != c) cas[i]->MarcaNoPosible();
}

Unico Fila::CandidatoUnico() {
    bool valor[9] = {false, false, false, false, false, false, false, false, false};
    int cuenta = 0;
    int i, c;

    // Contar valores, si son 8 devolver valor que falta:
    for(i=0; i<9; i++) {
        if(cas[i]->Valor()) {
            cuenta++;
            valor[cas[i]->Valor()-1] = true;
        } else
            c=i;
    }
    if(cuenta == 8) {
        i=0;
        while(valor[i++]);
        return Unico(c, i);
    }
    return Unico(-1,0);
}

int Fila::UnicoPosible() {
    int cuenta=0;
    int pos = -1;
    for(int i=0; i<9; i++) {
        if(cas[i]->Posible()) {
            cuenta++;
            pos = i;
        }
    }
    if(cuenta == 1) return pos;
    return -1;
}

// Definicin de clase Columna
void Columna::Iniciar(Tablero &t, int c) {
    for(int i = 0; i < 9; i++) cas[i] = &t.GetCasilla(i, c);
}

bool Columna::Presente(int n) {
    for(int i=0; i<9; i++) if(cas[i]->Valor() == n) return true;
    return false;
}

void Columna::Criba() {
    for(int i=0; i<9; i++) cas[i]->MarcaNoPosible();
}

void Columna::CribaMenosCaja(int f) {
    for(int i=0; i<9; i++) if(i/3 != f) cas[i]->MarcaNoPosible();
}

Unico Columna::CandidatoUnico() {
    bool valor[9] = {false, false, false, false, false, false, false, false, false};
    int cuenta = 0;
    int i, c;

    // Contar valores, si son 8 devolver valor que falta:
    for(i=0; i<9; i++) {
        if(cas[i]->Valor()) {
            cuenta++;
            valor[cas[i]->Valor()-1] = true;
        } else
            c=i;
    }
    if(cuenta == 8) {
        i=0;
        while(valor[i++]);
        return Unico(c, i);
    }
    return Unico(-1,0);
}

int Columna::UnicoPosible() {
    int cuenta=0;
    int pos = -1;
    for(int i=0; i<9; i++) {
        if(cas[i]->Posible()) {
            cuenta++;
            pos = i;
        }
    }
    if(cuenta == 1) return pos;
    return -1;
}

// Definicin de clase Caja
Caja::Caja() {
    for(int i=0; i < 9; i++) {
        cas[i] = NULL;
        for(int j=0; j<3; j++) {
            punteroFila[j][i] = punteroColumna[j][i] = 0;
        }
    }
}

void Caja::Iniciar(Tablero &t, int c) {
    for(int i = 0; i < 3; i++)
        for(int j = 0; j < 3; j++)
            cas[i*3+j] = &t.GetCasilla((c/3)*3+i, (c%3)*3+j);
}

bool Caja::Presente(int n) {
    for(int i=0; i<9; i++) if(cas[i]->Valor() == n) return true;
    return false;
}

void Caja::Criba() {
    for(int i=0; i<9; i++) cas[i]->MarcaNoPosible();
}

int Caja::CuentaPosibles() {
    int cuenta=0;
    for(int i=0; i<9; i++)
        if(cas[i]->Posible()) cuenta++;
    return cuenta;
}

int Caja::PrimeroPosible() {
    for(int i=0; i<9; i++)
        if(cas[i]->Posible()) return i;
    return -1;
}

int Caja::SegundoPosible(int p) {
    for(int i=p+1; i<9; i++)
        if(cas[i]->Posible()) return i;
    return -1;
}

int Caja::TerceroPosible(int s) {
    for(int i=s+1; i<9; i++)
        if(cas[i]->Posible()) return i;
    return -1;
}

bool Caja::AsignaPunteroFila(int f, int v, int n) {
    if(punteroFila[f][v-1] != n) {
        punteroFila[f][v-1] = n;
        return true;
    }
    return false;
}
bool Caja::AsignaPunteroColumna(int c, int v, int n) {
    if(punteroColumna[c][v-1] != n) {
        punteroColumna[c][v-1] = n;
        return true;
    }
    return false;
}

Unico Caja::CandidatoUnico() {
    bool valor[9] = {false, false, false, false, false, false, false, false, false};
    int cuenta = 0;
    int i, c;

    // Contar valores, si son 8 devolver valor que falta:
    for(i=0; i<9; i++) {
        if(cas[i]->Valor()) {
            cuenta++;
            valor[cas[i]->Valor()-1] = true;
        } else
            c=i;
    }
    if(cuenta == 8) {
        i=0;
        while(valor[i++]);
        return Unico(c, i);
    }
    return Unico(-1,0);
}

// Definicin de clase Tablero
Tablero::Tablero(char* t) : cuenta(0) {
    int f, c;

    for(f=0; f < 9; f++)
        for(c=0; c < 9; c++) {
            casilla[f][c] = Casilla(f, c, t[f*9+c]-'0', t[f*9+c] != '0');
        }
    for(f=0; f < 9; f++) fila[f].Iniciar(*this, f);
    for(c=0; c < 9; c++) columna[c].Iniciar(*this, c);
    for(c=0; c < 9; c++) caja[c].Iniciar(*this, c);
    flog = std::ofstream("solucion.log", std::ios::out);
}

void Tablero::Log(char *cad) {
    flog << cad << std::endl;
    LogPantalla(cad);
}

bool Tablero::Valido(int f, int c, int v) {
    return !fila[f].Presente(v) && !columna[c].Presente(v) && !caja[(f/3)*3+(c/3)].Presente(v);
}

bool Tablero::Soluciones() {
    int f, c, n;
    for(f=0; f<9; f++) {
        for(c=0; c<9; c++) {
            if(casilla[f][c].Valor() == 0) {
                for(n=1; n<=9; n++) {
                    if(Valido(f, c, n)) {
                        casilla[f][c].Asignar(n);
                        if(Soluciones()) return true;
                        else casilla[f][c].Asignar(0);
                    }
                }
                return false;
            }
        }
    }
    cuenta++;
    return false;
}

void Tablero::Asignar(int f, int c, int v) {
    casilla[f][c].Asignar(v);
    casilla[f][c].Resaltar();
    // Eliminar marcas de lapiz en fila, columna y caja:
    int k = (f/3)*3 + (c/3);
    for(int i=0; i<9; i++) {
        if(fila[f].Cas(i)->DesmarcaLapiz(v)) fila[f].Cas(i)->Resaltar();
        if(columna[c].Cas(i)->DesmarcaLapiz(v)) columna[c].Cas(i)->Resaltar();
        if(caja[k].Cas(i)->DesmarcaLapiz(v)) caja[k].Cas(i)->Resaltar();
    }
}

void Tablero::LimpiaResaltes() {
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            casilla[f][c].Resaltar(false);
        }
    }
}

void Tablero::LimpiaCriba() {
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            casilla[f][c].MarcaPosible();
        }
    }
}

void Tablero::Criba(int v) {
    int ncaja=0;

    LimpiaCriba();
    LimpiaResaltes();
    // Aplicar criba a valores finales:
    for(int i=0; i<9; i++) {
        if(fila[i].Presente(v)) fila[i].Criba();
        if(columna[i].Presente(v)) columna[i].Criba();
        if(caja[i].Presente(v)) caja[i].Criba();
    }
    // Eliminar de la criba el resto de casillas con valor final
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(casilla[f][c].Valor()) casilla[f][c].MarcaPosible(false);
        }
    }
    // Aplicar criba a parejas y trios punteros de cajas:
    // No desmarcar como posibles las casillas miembros de parejas y trios punteros:
    for(int f=0; f<3; f++) { // Fila de cajas
        for(int c=0; c<3; c++) { // Columna de cajas
            for(int i=0; i<3; i++) { // Fila de punteros
                if(caja[ncaja].PunteroFila(i, v)) fila[f*3+i].CribaMenosCaja(c);
            }
            for(int i=0; i<3; i++) { // Columna de punteros
                if(caja[ncaja].PunteroColumna(i, v)) columna[c*3+i].CribaMenosCaja(f);
            }
            ncaja++;
        }
    }
}

// TODO: Procesar criba por cajas, posiblemente tambin puede ser til hacerlo por
// filas y columnas. ???
bool Tablero::ProcesarCriba(int v) {
    bool retval = false;
    int p, s, t;
    int f, c;
    char cad[100];

    Texto(38, 1, "                                          ", 15);

    for(int i=0; i<9; i++) { // Verificar cada caja
        switch(caja[i].CuentaPosibles()) {
        case 1:
            // Final:
            p = caja[i].PrimeroPosible();
            f = caja[i].Cas(p)->Fila();
            c = caja[i].Cas(p)->Columna();
            Asignar(f, c, v);
            sprintf(cad, "Valor final %d en %c%c", v, 'A'+f, '1'+c);
            Log(cad);
            retval=true;
            break;
        case 2:
            // pareja puntero:
            p = caja[i].PrimeroPosible();
            s = caja[i].SegundoPosible(p);
            // Aadir marcas de lpiz y marcarlas en el tablero
            if(caja[i].Cas(p)->MarcaLapiz(v))
                caja[i].Cas(p)->Resaltar();
            if(caja[i].Cas(s)->MarcaLapiz(v))
                caja[i].Cas(s)->Resaltar();
            // Verificar si p y s estn alineados y guardar pareja puntero si lo estn
            if(caja[i].Cas(p)->Fila() == caja[i].Cas(s)->Fila()) {
                if(caja[i].AsignaPunteroFila(caja[i].Cas(p)->Fila()%3, v, 2)) {
                    sprintf(cad, "Par puntero caja %d fila %c, valor %d", 1+i, 'A'+caja[i].Cas(p)->Fila(), v);
                    Log(cad);
                    retval=true;
                }
            }
            if(caja[i].Cas(p)->Columna() == caja[i].Cas(s)->Columna()) {
                if(caja[i].AsignaPunteroColumna(caja[i].Cas(p)->Columna()%3, v, 2)) {
                    sprintf(cad, "Par puntero caja %d columna %c, valor %d", 1+i, '1'+caja[i].Cas(p)->Columna(), v);
                    Log(cad);
                    retval=true;
                }
            }
            break;
        case 3:
            // trio puntero:
            p = caja[i].PrimeroPosible();
            s = caja[i].SegundoPosible(p);
            t = caja[i].TerceroPosible(s);
            // Verificar si p, s y t estn alineados y guardar trie puntero si lo estn
            if(caja[i].Cas(p)->Fila() == caja[i].Cas(s)->Fila() && caja[i].Cas(p)->Fila() == caja[i].Cas(t)->Fila()) {
                if(caja[i].AsignaPunteroFila(caja[i].Cas(p)->Fila()%3, v, 3)) {
                    sprintf(cad, "Trio puntero caja %d columna %c, valor %d", 1+i, '1'+caja[i].Cas(p)->Columna(), v);
                    Log(cad);
                    retval=true;
                }
            }
            if(caja[i].Cas(p)->Columna() == caja[i].Cas(s)->Columna() && caja[i].Cas(p)->Columna() == caja[i].Cas(t)->Columna()) {
                if(caja[i].AsignaPunteroColumna(caja[i].Cas(p)->Columna()%3, v, 3)) {
                    sprintf(cad, "Trio puntero caja %d columna %c, valor %d", 1+i, '1'+caja[i].Cas(p)->Columna(), v);
                    Log(cad);
                    retval=true;
                }
            }
            break;
        }
    }
    return retval;
}

bool Tablero::CandidatoUnico() {
    bool retval = false;
    Unico u(0,0);
    char cad[100];

    // Buscar candidatos nicos en filas, columnas y cajas:
    for(int f=0; f<9; f++) {
        u = fila[f].CandidatoUnico();
        if(u.Casilla() != -1) {
            Asignar(f, u.Casilla(), u.Valor());
            sprintf(cad, "Candidato unico: Valor final %d en %c%c", u.Valor(), 'A'+f, '1'+u.Casilla());
            Log(cad);
            retval = true;
        }
    }

    for(int c=0; c<9; c++) {
        u = columna[c].CandidatoUnico();
        if(u.Casilla() != -1) {
            Asignar(u.Casilla(), c, u.Valor());
            sprintf(cad, "Candidato unico: Valor final %d en %c%c", u.Valor(), 'A'+u.Casilla(), '1'+c);
            Log(cad);
            retval = true;
        }
    }

    for(int i=0; i<9; i++) {
        u = caja[i].CandidatoUnico();
        if(u.Casilla() != -1) {
            Asignar(caja[i].Cas(u.Casilla())->Fila(), caja[i].Cas(u.Casilla())->Columna(), u.Valor());
            sprintf(cad, "Candidato unico: Valor final %d en %c%c", u.Valor(), 'A'+caja[i].Cas(u.Casilla())->Fila(), '1'+caja[i].Cas(u.Casilla())->Columna());
            Log(cad);
            retval = true;
        }
    }
    return retval;
}

bool Tablero::ProcesarCribaSolitarios(int v) {
    bool retval = false;
    int fil, col;
    char cad[100];

    for(int i=0; i<9; i++) { // Verificar cada fila
        col = fila[i].UnicoPosible();
        if(col != -1) {
            Asignar(i, col, v);
            sprintf(cad, "Candidato solitario: Valor final %d en %c%c", v, 'A'+i, '1'+col);
            Log(cad);
            retval = true;
        }
    }

    for(int i=0; i<9; i++) { // Verificar cada columna
        fil = columna[i].UnicoPosible();
        if(fil != -1) {
            Asignar(fil, i, v);
            retval = true;
            sprintf(cad, "Candidato solitario: Valor final %d en %c%c", v, 'A'+fil, '1'+i);
            Log(cad);
        }
    }
    return retval;
}

void Tablero::Mostrar() {
    int f, c;
    WORD color1, color2;

    color1 = 15;
    color2 = 3;

    for(f=0; f<9; f++) {
        for(c=0; c<9; c++) {
            if(f%3==0) {
                Texto(f*4,  c*4, "+", color1);
                Texto(f*4,  c*4+1, "---+", color1);
            } else {
                Texto(f*4,  c*4, "+", color2);
                Texto(f*4,  c*4+1, "---+", color2);
            }
            if(c%3==0) {
                Texto(f*4+1,c*4, "|", color1);
                Texto(f*4+2,c*4, "|", color1);
                Texto(f*4+3,c*4, "|", color1);
            } else {
                Texto(f*4+1,c*4, "|", color2);
                Texto(f*4+1,c*4+4, "|", color2);
                Texto(f*4+2,c*4, "|", color2);
                Texto(f*4+2,c*4+4, "|", color2);
                Texto(f*4+3,c*4, "|", color2);
                Texto(f*4+3,c*4+4, "|", color2);
            }
            Texto(f*4+1,c*4+4, "|", color1);
            Texto(f*4+2,c*4+4, "|", color1);
            Texto(f*4+3,c*4+4, "|", color1);
        }
    }
    for(c=0; c<9; c++) {
        Texto(36,  c*4, "+", color1);
        Texto(36,  c*4+1, "---+", color1);
    }

    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            casilla[f][c].Mostrar();
        }
    }
    Texto(40, 1, " ", color1);
}

int main()
{
    bool repetir;
    bool insistir;
    char cad[100];

    Tablero T(tablero);
    T.Mostrar();

    T.Soluciones();
    // Si la solucin no es nica, no intentar resolver
    if(T.Cuenta() != 1) {
        if(T.Cuenta() == 0)
            Texto(40, 0, "El sudoku no tiene solucin", 15);
        else
            Texto(40, 0, "El sudoku tiene varias soluciones", 15);
        return 0;
    }

    do {
        repetir=false;
        for(int v=1; v<=9; v++) {
            sprintf(cad, "Criba con valor %d", v);
            Texto(1, 40, cad, 7);
            do {
                T.Criba(v);
                insistir = T.ProcesarCriba(v);
                repetir |= insistir;
                T.Mostrar();
//                Pausa();
            } while(insistir);
        }
        if(!repetir && T.CandidatoUnico()) {
            repetir = true;
            Texto(2,40,"ProcesarCandidatoUnico", 7);
            T.Mostrar();
            Pausa();
        }
        if(!repetir) {
            for(int v=1; v<=9; v++) {
                sprintf(cad, "Criba con valor %d", v);
                Texto(1, 40, cad, 7);
                T.Criba(v);
                repetir |= T.ProcesarCribaSolitarios(v);
                if(repetir) {
                    Texto(3,40,"ProcesarCribaSolitarios", 7);
                    T.Mostrar();
                    Pausa();
                }
            }
        }
    } while(repetir);


    T.LimpiaCriba();
    T.LimpiaResaltes();
    T.Mostrar();
    return 0;
}
